<svg xmlns="http://www.w3.org/2000/svg" width="{{ left.width+config.left_pad + right.width+config.right_pad }}" height="20">
  <rect width="{{ (left.width+config.left_pad) + (right.width+config.right_pad)}}" height="20" fill="{{ left.color }}"/>
  <rect x="{{ left.width+config.left_pad}}" width="{{ right.width + config.right_pad}}" height="20" fill="{{ right.color }}"/>
  <g fill="#fff" text-anchor="middle" font-family="{{config.font_face}},Verdana,Geneva,sans-serif" font-size="{{config.font_size}}">
    <text x="{{ (left.width+config.left_pad)/2 }}" y="14">{{ left.text }}</text>
    <text x="{{ (left.width+config.left_pad)+(right.width+config.right_pad)/2 }}" y="14">{{ right.text }}</text>
  </g>
</svg>
