.bb-settings-view {
  .card {
    margin-bottom: 1.25rem;
   }

  .card-header {
    margin-bottom: 0;
    padding-top: 0.625rem;
    padding-bottom: 0.625rem;
  }

  .card-title {
    margin-bottom: 0;
  }

  .h5 {
    font-size: 1rem;
  }

  .card-body {
    padding: 0.5rem;
  }
}
