$bb-color-results-success: #000;
$bb-color-background-results-success: #8d4;
$bb-color-border-results-success: #4F8530;

$bb-color-results-warnings: #FFFFFF;
$bb-color-background-results-warnings: #fa3;
$bb-color-border-results-warnings: #C29D46;

$bb-color-results-failure: #000;
$bb-color-background-results-failure: #e88;
$bb-color-border-results-failure: #A77272;

$bb-color-results-skipped: #000;
$bb-color-background-results-skipped: #AADDEE;
$bb-color-border-results-skipped: #AADDEE;

$bb-color-results-exception: #FFFFFF;
$bb-color-background-results-exception: #c6c;
$bb-color-border-results-exception: #ACA0B3;

$bb-color-results-retry: #000;
$bb-color-background-results-retry: #ecc;
$bb-color-border-results-retry: #ACA0B3;

$bb-color-results-cancelled: #000;
$bb-color-background-results-cancelled: #ecc;
$bb-color-border-results-cancelled: #ACA0B3;

$bb-color-results-pending: #000;
$bb-color-background-results-pending: #E7D100;
$bb-color-border-results-pending: #ACA0B3;

$bb-color-results-unknown: #000;
$bb-color-background-results-unknown: #EEE;
$bb-color-border-results-unknown: #ACA0B3;
