@import "bootstrap/scss/bootstrap";
@import '../../styles/colors.scss';

.bb-badge-round {
  display: inline-block;
  min-width: 10px;
  padding: 3px 7px;
  font-size: $font-size-sm;
  font-weight: $badge-font-weight;
  line-height: $font-size-sm;
  vertical-align: baseline;
  white-space: nowrap;
  text-align: center;
  border-radius: 10px;
  transition: all 100ms cubic-bezier(0.25, 0.25, 0.75, 0.75);

  .badge-active {
    display: none;
  }
  &:hover,
  &:focus {
    -webkit-transform: scale(1.3);
    transform: scale(1.3);
    .badge-active {
      display: inline-block;
    }
    .badge-inactive {
      display: none;
    }
  }

  &.results_SUCCESS, &.worker_CONNECTED {
    color: $bb-color-results-success;
    background-color: $bb-color-background-results-success;
    border-color: $bb-color-border-results-success;
  }

  &.results_WARNINGS {
    color: $bb-color-results-warnings;
    background-color: $bb-color-background-results-warnings;
    border-color: $bb-color-border-results-warnings;
  }

  &.results_FAILURE, &.worker_DISCONNECTED {
    color: $bb-color-results-failure;
    background-color: $bb-color-background-results-failure;
    border-color: $bb-color-border-results-failure;
  }

  &.results_SKIPPED {
    color: $bb-color-results-skipped;
    background: $bb-color-background-results-skipped;
    border-color: $bb-color-border-results-skipped;
  }

  &.results_EXCEPTION {
    color: $bb-color-results-exception;
    background-color: $bb-color-background-results-exception;
    border-color: $bb-color-border-results-exception;
  }

  &.results_RETRY {
    color: $bb-color-results-retry;
    background-color: $bb-color-background-results-retry;
    border-color: $bb-color-border-results-retry;
  }

  &.results_CANCELLED {
    color: $bb-color-results-cancelled;
    background-color: $bb-color-background-results-cancelled;
    border-color: $bb-color-border-results-cancelled;
  }

  &.results_PENDING {
    color: $bb-color-results-pending;
    background-color: $bb-color-background-results-pending;
    border-color: $bb-color-border-results-pending;
  }

  &.results_UNKNOWN {
    color: $bb-color-results-unknown;
    background-color: $bb-color-background-results-unknown;
    border-color: $bb-color-border-results-unknown;
  }
}
