.changedetails {
  width: 100%;
}

.changedetails-heading {
  width: 100%;
}

.changedetails-heading > * {
  margin-right: 0.5em;
}

.changedetails-properties {
  padding: unset;
  margin: unset;
  border: unset;
  background-color: unset
}
