.bb-build-view-pager {
  text-align: center;
  padding: 0;
  margin-bottom: 30px;

  li {
    display: inline;
  }

  li.previous {
    float: left;
  }

  li.next {
    float: right;
  }
}

.bb-build-view {
  .tab-content {
    width: 100%;
  }
}

.bb-build-view-nav-button {
  padding: 5px 14px;
  background-color: #fff;
  border: 1px solid #ddd;
  border-radius: 15px;
  display: block;
}
