
.bb-buildsummary-tooltip-collapsed-entries {
  font-size: 1.333em;
  line-height: 0.75em;
  vertical-align: -15%;
}

.bb-buildsummary-tooltip-step-badge, .bb-buildsummary-tooltip-step-name {
  float: left;
}

.bb-buildsummary-tooltip-step-time {
  float: right;
}
