.bb-raw-data-key-value {
  & > dt {
    float: left;
    width: 160px;
    clear: left;
    text-align: right;
    overflow: hidden;
    text-overflow: ellipsis;
    white-space: nowrap;
  }

  & > dd {
    text-align: left;
    padding-left: 20px;
    width: 70%;
    display: inline-block;
    box-sizing: border-box;
  }
}
