.bb-projects-view-form-control {
  display: block;
  width: 100%;
  height: 34px;
  padding: 6px 12px;
  margin-bottom: 20px;
  font-size: 14px;
  line-height: 1.42857143;
  color: #555555;
  background-color: #fff;
  border: 1px solid #ccc;
  border-radius: 4px;
  box-shadow: inset 0 1px 1px rgba(0, 0, 0, 0.075);
  transition: border-color ease-in-out .15s, box-shadow ease-in-out .15s;
}

.bb-projects-view-list {
  & > a:hover > li {
    background-color: #dddddd;
  }
  & > a > li {
    padding: 0.5rem 0.75rem;
    margin: 0.25rem 0;
  }
}

.bb-projects-view-container {
  padding-right: 15px;
  padding-left: 15px;
  margin-left: 0;
  margin-right: 0;
}
