.fade.modal.show {
  // Otherwise the modal is not shown.
  opacity: 1;
}

@media screen {
  .bb-forcebuild-modal {
    .modal-dialog {
      max-width: 100%;
      width: 90%;
    }

    .modal-body {
      overflow-y: auto;
      max-height: 80vh;
    }
  }
}

.tooltip-icon {
  margin-left: 5px;
  cursor: pointer;
  position: relative;
}