@import "bootstrap/scss/bootstrap";

.bb-build-summary {
  .card-header {
    color: #333333;
    background-color: #f5f5f5;
    border-color: #ddd;
    padding: 0.375rem 1rem;
    border-bottom: 1px solid transparent;
    border-top-left-radius: 3px;
    border-top-right-radius: 3px;
  }

  .bb-build-summary-details,
  .bb-build-summary-step-details {
    float: right;
  }

  .bb-build-request-counter {
    color: $gray-600;
    font-style: italic;
  }

  .btn-xs {
    padding: 1px 5px;
    font-size: 12px;
    line-height: 1.5;
    border-radius: 3px;
  }

  .btn-default {
    color: #333;
    background-color: #fff;
    border-color: #ccc;
  }

  .group-results > li .bb-build-request-summary > div {
    margin-bottom: 0px;
    margin-top: 0px;
  }
}

.bb-build-summary-step-line {
  position: relative;

  &.list-group-item {
    padding: 0.375rem 1rem;
  }

  &.list-group-item.bb-anchor-target {
    border: 2px solid #ffff00;
  }
  .bb-build-summary-step-anchor-link {
    display: none;
    color: $gray-500;
    position: absolute;
    left: 0.5em;
  }
}

.bb-build-summary-step-line:hover {
  .bb-build-summary-step-anchor-link {
    display: inline-block;
  }
}
