.bb-tag-filter-manager-tag {
  padding: 0.2em 0.6em 0.3em 0.6em;
  font-size: 75%;
  font-weight: 700;
  line-height: 1;
  color: #fff;
  background-color: #777777;
  text-align: center;
  white-space: nowrap;
  border-radius: 0.25em;
}

.bb-tag-filter-manager-tag-filtered {
  background-color: #5cb85c;
}

#bb-tag-filter-manager-help-popover {
  padding: 1px;
  border-radius: 6px;
  box-shadow: 0 5px 10px rgba(0, 0, 0, 0.2);
  margin-top: 10px;

  .popover-header {
    padding: 8px 14px;
    margin: 0;
    font-size: 14px;
    background-color: #f7f7f7;
    border-bottom: 1px solid #ebebeb;
    border-radius: 5px 5px 0 0;
  }

  .popover-body {
    padding: 9px 14px;

    pre {
      display: block;
      padding: 9.5px;
      margin: 0 0 10px;
      font-size: 13px;
      word-break: break-all;
      word-wrap: break-word;
      background-color: #f5f5f5;
      border: 1px solid #ccc;
      border-radius: 4px;
    }
  }
}
