@import "bootstrap/scss/bootstrap";
@import '../../styles/colors.scss';

.bb-log-search-field {
  position: relative;
  width: 300px;
  display: inline-block;
  margin-right: 0.5rem;
}

.bb-log-search-field-icon {
  position: absolute;
  top: 0.7rem;
  margin-left: 0.5rem;
  color: $body-color;
}

.bb-log-search-field-nav {
  position: absolute;
  right: 0.2rem;
  top: 0;
  display: flex;
  flex-wrap: nowrap;
  color: $body-color;
}

.bb-log-search-field-result-count {
  padding: 0.5rem;
}

.bb-log-search-field-options {
  position: absolute;
  top: 0.4rem;
  color: $body-color;
  padding: 0;
  border: none;
  background: none;
}

.bb-log-search-field-options-toggled {
  border: rgb(0, 0, 255);
  background-color: rgba(0, 0, 255, 0.4);
}
.bb-log-search-field-options-untoggled {
  border: none;
  background-color: rgba(255, 255, 255, 0);
}

.bb-log-search-field-options-case {
  margin-left: 1.75rem;
}

.bb-log-search-field-options-regex {
  margin-left: 2.75rem;
}

.bb-log-search-field-nav-button {
  border-radius: 0.25rem;
  border: 0px solid transparent;
  background-color: transparent;
  margin: 0.4rem 0.2rem;
  width: 1.75rem;
  height: 1.75rem;
  font-size: 0.8rem;
}

.bb-log-search-field-nav-button:hover {
  background-color: $bb-color-button-hover-light;
}

.bb-log-search-field-text {
  width: 100%;
  padding: 0.375rem 0.75rem 0.375rem 4rem;
  border-radius: 0.25rem;
  overflow: visible;
}
