.bb-home-view {
  .card-header {
    a {
      color: white;
    }
    background-color: #337ab7;
    border-color: #337ab7;
    padding: 10px 15px;
  }

  .card-body {
    padding: 0.5rem;
  }

  .bb-home-builder-card {
    border-color: #337ab7;
  }
}
